@@ -1,27 +1,16 @@
 from collections.abc import Sequence
-from enum import Enum
 import logging
 from pathlib import Path
 from typing import Mapping
-from PIL import Image, ImageDraw, ImageFont
+from PIL import Image
 from skit._types import Real, LayoutDef, Color, FreeTypeFont, Alignment
+from skit.render import DrawCommand, SingleImageRenderer
 from abc import ABC, abstractmethod
 
 
 logger = logging.getLogger(__file__)
 
 
-# some defaults for fallback
-_DEFAULT_FONT = ImageFont.truetype('Helvetica', 16)
-_DEFAULT_COLOR = 'black'
-_DEFAULT_THICKNESS = 1
-
-class _DrawCommand(Enum):
-    TEXT = 'text'
-    RECTANGLE = 'rect'
-    IMAGE = 'image'
-
-
 class CardManipulation(ABC):
     @abstractmethod
     def background(self, color: str): pass
@@ -68,10 +57,14 @@ class CardManipulation(ABC):
     @abstractmethod
     def render_png(self, filename: str): pass
 
+    @abstractmethod
+    def render_pdf(self, filename: str, resolution: int, single_file: bool): pass
+
 
 class Card(CardManipulation):
     "A single card."
     def __init__(self, width: int = 750, height: int = 1050):
+        """Create a new card of a certain `width` and `height`."""
         self._width = width
         self._height = height
         self._layouts = {}
@@ -98,6 +91,7 @@ class Card(CardManipulation):
             'height': layoutdef.height,
             'h_align': layoutdef.h_align,
             'v_align': layoutdef.v_align,
+            'scale': layoutdef.scale,
         }
     
     def layouts(self, names: Sequence[str], layoutdefs: Sequence[LayoutDef]):
@@ -119,7 +113,7 @@ class Card(CardManipulation):
         if layout in self._layouts:
             logger.debug(f"adding '{text}' in {layout}")
             self._commands.append({
-                'op': _DrawCommand.TEXT,
+                'op': DrawCommand.TEXT,
                 'layout': layout,
                 'text': text,
                 'font': font,
@@ -133,7 +127,7 @@ class Card(CardManipulation):
         if layout in self._layouts:
             logger.debug(f"adding rectangle for {layout}")
             self._commands.append({
-                'op': _DrawCommand.RECTANGLE,
+                'op': DrawCommand.RECTANGLE,
                 'layout': layout,
                 'color': color,
                 'thickness': thickness,
@@ -147,7 +141,7 @@ class Card(CardManipulation):
         if layout in self._layouts:
             logger.debug(f"adding rectangle for {layout}")
             self._commands.append({
-                'op': _DrawCommand.RECTANGLE,
+                'op': DrawCommand.RECTANGLE,
                 'layout': layout,
                 'color': color,
                 'thickness': 1,
@@ -161,7 +155,7 @@ class Card(CardManipulation):
         if layout in self._layouts:
             logger.debug(f"adding image for {layout}")
             self._commands.append({
-                'op': _DrawCommand.IMAGE,
+                'op': DrawCommand.IMAGE,
                 'layout': layout,
                 'image': image,
             })
@@ -172,98 +166,30 @@ class Card(CardManipulation):
         "Render this card as a PNG."
         logger.debug(f"rendering {filename}")
 
-        with Image.new('RGBA', (self._width, self._height), self._background) as im:
-            d = ImageDraw.Draw(im)
-
-            for cmd in self._commands:
-                op = cmd.pop('op')
-                match op:
-                    case _DrawCommand.TEXT:
-                        self._render_png_text(d, **cmd)
-                    case _DrawCommand.RECTANGLE:
-                        self._render_png_rectangle(d, **cmd)
-                    case _DrawCommand.IMAGE:
-                        self._render_png_image(im, **cmd)
-                    case _:
-                        raise ValueError(cmd)
-
-            im.save(filename)
-
-    def _render_png_text(self, d, layout, text, color, font):
-        logger.debug(f"rendering text '{text}' at {layout}")
-        layout = self._layouts[layout]
-
-        match layout['h_align']:
-            case Alignment.BEGIN:
-                anchor_h = 'l'
-                x = layout['x']
-            case Alignment.MIDDLE:
-                anchor_h = 'm'
-                x = layout['x'] + (layout['width'] // 2)
-            case Alignment.END:
-                anchor_h = 'r'
-                x = layout['x'] + layout['width']
-            case _:
-                raise ValueError(f"h_align value '{layout['h_align']}' unrecognized")
-
-        match layout['v_align']:
-            case Alignment.BEGIN:
-                y = layout['y']
-                anchor_v = 'a'
-            case Alignment.MIDDLE:
-                y = layout['y'] + (layout['height'] // 2)
-                anchor_v = 'm'
-            case Alignment.END:
-                y = layout['y'] + layout['height']
-                anchor_v = 'd'
-            case _:
-                raise ValueError(f"v_align value '{layout['v_align']}' unrecognized")
-
-        d.text(
-            [x, y],
-            text,
-            fill=color if color else _DEFAULT_COLOR,
-            font=font if font else _DEFAULT_FONT,
-            anchor=f"{anchor_h}{anchor_v}",
+        im = (
+            SingleImageRenderer(self._layouts)
+            .render(self._width, self._height, self._background, self._commands)
         )
-    
-    def _render_png_rectangle(self, d, layout, color, thickness, filled):
-        logger.debug(f"rendering rectangle on {layout}")
-        layout = self._layouts[layout]
-        d.rectangle(
-            [
-                layout['x'],
-                layout['y'],
-                layout['x'] + layout['width'],
-                layout['y'] + layout['height'],
-            ],
-            fill=color if filled and color else None,
-            outline=color if color else _DEFAULT_COLOR,
-            width=thickness if thickness else _DEFAULT_THICKNESS,
+        im.save(filename, format='PNG')
+
+    def render_pdf(self, filename: str, resolution: int, single_file=True):
+        """
+        Render this card as a PDF. It makes no sense to render a single
+        card to multiple files, so `single_file` is ignored. PDFs don't
+        support the alpha channel, so remove it.
+        """
+        logger.debug(f"rendering {filename}")
+
+        im = self._get_rgb_image_for_pdf(resolution)
+        im.save(filename, format='PDF', resolution=resolution)
+
+    def _get_rgb_image_for_pdf(self, resolution: int):
+        logging.debug(f"rendering RGB image")
+
+        im = (
+            SingleImageRenderer(self._layouts)
+            .render(self._width, self._height, self._background, self._commands)
         )
-
-    def _render_png_image(self, im, layout, image):
-        logger.debug(f"rendering image at {layout}")
-        layout = self._layouts[layout]
-        with Image.open(image) as art:
-            match layout['h_align']:
-                case Alignment.BEGIN:
-                    left = layout['x']
-                case Alignment.MIDDLE:
-                    left = layout['x'] + (layout['width'] - art.width) // 2
-                case Alignment.END:
-                    left = layout['x'] + layout['width'] - art.width
-                case _:
-                    raise ValueError(f"h_align value '{layout['h_align']}' unrecognized")
-
-            match layout['v_align']:
-                case Alignment.BEGIN:
-                    top = layout['y']
-                case Alignment.MIDDLE:
-                    top = layout['y'] + (layout['height'] - art.height) // 2
-                case Alignment.END:
-                    top = layout['y'] + layout['height'] - art.height
-                case _:
-                    raise ValueError(f"v_align value '{layout['v_align']}' unrecognized")
-
-            im.alpha_composite(art, (left, top))
+        final = Image.new('RGB', im.size, self._background)
+        final.paste(im)
+        return final
